from typing import (
    List,
    Optional,
)

from fdk_rdf_parser.classes import (
    CVContactPoint,
    Evidence,
    OpeningHoursSpecification,
    Output,
    Service,
)
from .reference_data import PublicServiceReferenceData
from .utils import (
    collect_eu_data_themes,
    collect_los_themes,
    extend_skos_code,
    extend_skos_code_list,
)


def extend_cpsvno_service_with_reference_data(
    cpsvno_service: Service, ref_data: PublicServiceReferenceData
) -> Service:
    cpsvno_service.language = extend_skos_code_list(
        cpsvno_service.language, ref_data.linguisticsystem
    )
    cpsvno_service.contactPoint = extend_cv_contact_points(
        cpsvno_service.contactPoint, ref_data
    )
    cpsvno_service.produces = extend_cv_output(cpsvno_service.produces, ref_data)
    cpsvno_service.admsStatus = extend_skos_code(
        cpsvno_service.admsStatus, ref_data.statuses
    )
    cpsvno_service.hasInput = extend_cv_evidence(cpsvno_service.hasInput, ref_data)
    cpsvno_service.dctType = extend_skos_code_list(
        cpsvno_service.dctType, ref_data.types
    )

    cpsvno_service.losThemes = collect_los_themes(
        cpsvno_service.thematicAreaUris, ref_data.los_themes
    )
    cpsvno_service.euDataThemes = collect_eu_data_themes(
        cpsvno_service.thematicAreaUris, ref_data.eu_data_themes
    )

    return cpsvno_service


def extend_cv_contact_points(
    contact_points: Optional[List[CVContactPoint]], ref_data: PublicServiceReferenceData
) -> Optional[List[CVContactPoint]]:
    if contact_points is None:
        return contact_points

    for contact_point in contact_points:
        contact_point.language = extend_skos_code_list(
            contact_point.language, ref_data.linguisticsystem
        )
        contact_point.hoursAvailable = extend_opening_hours_spec(
            contact_point.hoursAvailable, ref_data=ref_data
        )
        contact_point.specialOpeningHours = extend_opening_hours_spec(
            contact_point.specialOpeningHours, ref_data=ref_data
        )
    return contact_points


def extend_opening_hours_spec(
    opening_hours: Optional[List[OpeningHoursSpecification]],
    ref_data: PublicServiceReferenceData,
) -> Optional[List[OpeningHoursSpecification]]:
    if opening_hours is None:
        return opening_hours

    for item in opening_hours:
        item.dayOfWeek = extend_skos_code_list(item.dayOfWeek, ref_data.week_days)
    return opening_hours


def extend_cv_output(
    outputs: Optional[List[Output]], ref_data: PublicServiceReferenceData
) -> Optional[List[Output]]:
    if outputs is None:
        return outputs

    for output in outputs:
        output.language = extend_skos_code_list(
            output.language, ref_data.linguisticsystem
        )
    return outputs


def extend_cv_evidence(
    evidences: Optional[List[Evidence]], ref_data: PublicServiceReferenceData
) -> Optional[List[Evidence]]:
    if evidences is None:
        return None

    for evidence in evidences:
        evidence.dctType = extend_skos_code_list(
            evidence.dctType, ref_data.evidence_type
        )
        evidence.language = extend_skos_code_list(
            evidence.language, ref_data.linguisticsystem
        )
    return evidences



from dataclasses import dataclass
from typing import (
    Any,
    List,
    Optional,
    Set,
)

from .catalog import Catalog
from .dcat_resource import PartialDcatResource
from .harvest_meta_data import HarvestMetaData
from .media_type import MediaTypeOrExtent
from .skos_code import ReferenceDataCode
from .skos_concept import SkosConcept


@dataclass
class DataService(PartialDcatResource):

    id: Optional[str] = None
    harvest: Optional[HarvestMetaData] = None
    endpointDescription: Optional[Set[str]] = None
    endpointURL: Optional[Set[str]] = None
    mediaType: Optional[List[ReferenceDataCode]] = None
    fdkFormat: Optional[List[MediaTypeOrExtent]] = None
    servesDataset: Optional[Set[str]] = None
    conformsTo: Optional[List[SkosConcept]] = None
    page: Optional[List[str]] = None
    catalog: Optional[Catalog] = None
    type: str = "dataservices"

    def add_values_from_dcat_resource(self: Any, values: PartialDcatResource) -> Any:
        self.identifier = values.identifier
        self.publisher = values.publisher
        self.title = values.title
        self.description = values.description
        self.descriptionFormatted = values.descriptionFormatted
        self.uri = values.uri
        self.accessRights = values.accessRights
        self.theme = values.theme
        self.keyword = values.keyword
        self.contactPoint = values.contactPoint
        self.dctType = values.dctType
        self.issued = values.issued
        self.modified = values.modified
        self.landingPage = values.landingPage
        self.language = values.language



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    OWL,
    RDFS,
)

from fdk_rdf_parser.classes import DctStandard
from fdk_rdf_parser.rdf_utils import (
    object_value,
    resource_list,
    value_list,
    value_translations,
)


def extract_dct_standard_list(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[DctStandard]]:
    values = []
    for resource in resource_list(graph, subject, predicate):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None
        values.append(
            DctStandard(
                uri=resource_uri,
                title=value_translations(graph, resource, DCTERMS.title),
                seeAlso=value_list(graph, resource, RDFS.seeAlso),
                versionInfo=object_value(graph, resource, OWL.versionInfo),
            )
        )

    return values if len(values) > 0 else None



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)

from fdk_rdf_parser.classes import (
    EuDataTheme,
    LosNode,
    MediaTypeOrExtent,
    MediaTypeOrExtentType,
    ReferenceDataCode,
)
from .reference_data_client import get_reference_data
from .utils import remove_scheme_and_trailing_slash


@dataclass
class DataServiceReferenceData:
    media_types: Optional[Dict[str, MediaTypeOrExtent]] = None


@dataclass
class DatasetReferenceData:
    provenancestatement: Optional[Dict[str, ReferenceDataCode]] = None
    rightsstatement: Optional[Dict[str, ReferenceDataCode]] = None
    frequency: Optional[Dict[str, ReferenceDataCode]] = None
    linguisticsystem: Optional[Dict[str, ReferenceDataCode]] = None
    referencetypes: Optional[Dict[str, ReferenceDataCode]] = None
    openlicenses: Optional[Dict[str, ReferenceDataCode]] = None
    location: Optional[Dict[str, ReferenceDataCode]] = None
    eu_data_themes: Optional[Dict[str, EuDataTheme]] = None
    los_themes: Optional[Dict[str, LosNode]] = None
    media_types: Optional[Dict[str, MediaTypeOrExtent]] = None


@dataclass
class InformationModelReferenceData:
    linguisticsystem: Optional[Dict[str, ReferenceDataCode]] = None
    openlicenses: Optional[Dict[str, ReferenceDataCode]] = None
    location: Optional[Dict[str, ReferenceDataCode]] = None
    eu_data_themes: Optional[Dict[str, EuDataTheme]] = None
    los_themes: Optional[Dict[str, LosNode]] = None


@dataclass
class PublicServiceReferenceData:
    linguisticsystem: Optional[Dict[str, ReferenceDataCode]] = None
    week_days: Optional[Dict[str, ReferenceDataCode]] = None
    statuses: Optional[Dict[str, ReferenceDataCode]] = None
    types: Optional[Dict[str, ReferenceDataCode]] = None
    evidence_type: Optional[Dict[str, ReferenceDataCode]] = None
    eu_data_themes: Optional[Dict[str, EuDataTheme]] = None
    los_themes: Optional[Dict[str, LosNode]] = None


def get_data_service_reference_data() -> DataServiceReferenceData:
    return DataServiceReferenceData(media_types=get_and_map_media_types())


def get_dataset_reference_data() -> DatasetReferenceData:
    return DatasetReferenceData(
        provenancestatement=get_and_map_provenance_statements(),
        rightsstatement=get_and_map_access_rights(),
        frequency=get_and_map_frequencies(),
        linguisticsystem=get_and_map_linguistic_system(),
        referencetypes=get_and_map_reference_types(),
        openlicenses=get_and_map_open_licenses(),
        location=get_and_map_locations(),
        eu_data_themes=get_and_map_eu_data_themes(),
        los_themes=get_and_map_los_themes(),
        media_types=get_and_map_media_types(),
    )


def get_info_model_reference_data() -> InformationModelReferenceData:
    return InformationModelReferenceData(
        linguisticsystem=get_and_map_linguistic_system(),
        openlicenses=get_and_map_open_licenses(),
        location=get_and_map_locations(),
        eu_data_themes=get_and_map_eu_data_themes(),
        los_themes=get_and_map_los_themes(),
    )


def get_public_service_reference_data() -> PublicServiceReferenceData:
    return PublicServiceReferenceData(
        linguisticsystem=get_and_map_linguistic_system(),
        week_days=get_and_map_week_days(),
        statuses=get_and_map_statuses(),
        evidence_type=get_and_map_evidence_types(),
        types=get_and_map_dct_types(),
        eu_data_themes=get_and_map_eu_data_themes(),
        los_themes=get_and_map_los_themes(),
    )


def parse_reference_codes(
    codes: Optional[List[Dict]],
) -> Optional[Dict[str, ReferenceDataCode]]:
    if codes is not None:
        return {
            remove_scheme_and_trailing_slash(str(code.get("uri"))): ReferenceDataCode(
                uri=str(code.get("uri")) if code.get("uri") is not None else None,
                code=str(code.get("code")) if code.get("code") is not None else None,
                prefLabel=code.get("label") if code.get("label") is not None else None,
            )
            for code in codes
        }
    else:
        return None


def parse_locations(
    nasjoner: Optional[List[Dict]],
    fylker: Optional[List[Dict]],
    kommuner: Optional[List[Dict]],
) -> Optional[Dict[str, ReferenceDataCode]]:
    locations = dict()
    if nasjoner is not None:
        for nasjon in nasjoner:
            locations[
                remove_scheme_and_trailing_slash(str(nasjon.get("uri")))
            ] = ReferenceDataCode(
                uri=str(nasjon.get("uri")) if nasjon.get("uri") is not None else None,
                code=str(nasjon.get("nasjonsnummer"))
                if nasjon.get("nasjonsnummer") is not None
                else None,
                prefLabel={"no": str(nasjon["nasjonsnavn"])}
                if nasjon.get("nasjonsnavn") is not None
                else None,
            )
    if fylker is not None:
        for fylke in fylker:
            locations[
                remove_scheme_and_trailing_slash(str(fylke.get("uri")))
            ] = ReferenceDataCode(
                uri=str(fylke.get("uri")) if fylke.get("uri") is not None else None,
                code=str(fylke.get("fylkesnummer"))
                if fylke.get("fylkesnummer") is not None
                else None,
                prefLabel={"no": str(fylke["fylkesnavn"])}
                if fylke.get("fylkesnavn") is not None
                else None,
            )
    if kommuner is not None:
        for kommune in kommuner:
            locations[
                remove_scheme_and_trailing_slash(str(kommune.get("uri")))
            ] = ReferenceDataCode(
                uri=str(kommune.get("uri")) if kommune.get("uri") is not None else None,
                code=str(kommune.get("kommunenummer"))
                if kommune.get("kommunenummer") is not None
                else None,
                prefLabel={"no": str(kommune["kommunenavnNorsk"])}
                if kommune.get("kommunenavnNorsk") is not None
                else None,
            )

    return locations if len(locations) > 0 else None


def get_and_map_locations() -> Optional[Dict[str, ReferenceDataCode]]:
    nasjoner = get_reference_data("/geonorge/administrative-enheter/nasjoner").get(
        "nasjoner"
    )
    fylker = get_reference_data("/geonorge/administrative-enheter/fylker").get("fylker")
    kommuner = get_reference_data("/geonorge/administrative-enheter/kommuner").get(
        "kommuner"
    )
    return parse_locations(nasjoner, fylker, kommuner)


def get_and_map_frequencies() -> Optional[Dict[str, ReferenceDataCode]]:
    frequencies = get_reference_data("/eu/frequencies").get("frequencies")
    return parse_reference_codes(frequencies)


def get_and_map_provenance_statements() -> Optional[Dict[str, ReferenceDataCode]]:
    provenance_statements = get_reference_data("/provenance-statements").get(
        "provenanceStatements"
    )
    return parse_reference_codes(provenance_statements)


def get_and_map_reference_types() -> Optional[Dict[str, ReferenceDataCode]]:
    reference_types = get_reference_data("/reference-types").get("referenceTypes")
    return parse_reference_codes(reference_types)


def get_and_map_open_licenses() -> Optional[Dict[str, ReferenceDataCode]]:
    open_licenses = get_reference_data("/open-licenses").get("openLicenses")
    return parse_reference_codes(open_licenses)


def get_and_map_linguistic_system() -> Optional[Dict[str, ReferenceDataCode]]:
    linguistic_systems = get_reference_data("/linguistic-systems").get(
        "linguisticSystems"
    )
    return parse_reference_codes(linguistic_systems)


def get_and_map_access_rights() -> Optional[Dict[str, ReferenceDataCode]]:
    access_rights = get_reference_data("/eu/access-rights").get("accessRights")
    return parse_reference_codes(access_rights)


def get_and_map_week_days() -> Optional[Dict[str, ReferenceDataCode]]:
    week_days = get_reference_data("/schema/week-days").get("weekDays")
    return parse_reference_codes(week_days)


def get_and_map_statuses() -> Optional[Dict[str, ReferenceDataCode]]:
    statuses = get_reference_data("/adms/statuses").get("statuses")
    return parse_reference_codes(statuses)


def get_and_map_dct_types() -> Optional[Dict[str, ReferenceDataCode]]:
    dct_types = get_reference_data("/eu/main-activities").get("mainActivities")
    return parse_reference_codes(dct_types)


def get_and_map_evidence_types() -> Optional[Dict[str, ReferenceDataCode]]:
    evidence_types = get_reference_data("/digdir/evidence-types").get("evidenceTypes")
    return parse_reference_codes(evidence_types)


def get_and_map_eu_data_themes() -> Optional[Dict[str, EuDataTheme]]:
    mapped = {}
    themes = get_reference_data("/eu/data-themes").get("dataThemes")
    if themes is not None:
        for theme in themes:
            eu_theme = EuDataTheme()
            eu_theme.add_values_from_dict(theme)
            mapped[remove_scheme_and_trailing_slash(str(theme.get("uri")))] = eu_theme
    return mapped if len(mapped) > 0 else None


def get_and_map_los_themes() -> Optional[Dict[str, LosNode]]:
    mapped = {}
    los_nodes = get_reference_data("/los/themes-and-words").get("losNodes")
    if los_nodes is not None:
        for los_node in los_nodes:
            los_theme = LosNode()
            los_theme.add_values_from_dict(los_node)
            mapped[
                remove_scheme_and_trailing_slash(str(los_node.get("uri")))
            ] = los_theme
    return mapped if len(mapped) > 0 else None


def get_and_map_media_types() -> Optional[Dict[str, MediaTypeOrExtent]]:
    media_types = {}
    iana_codes = get_reference_data("/iana/media-types").get("mediaTypes")
    if iana_codes is not None:
        for code in iana_codes:
            media_type_uri = str(code["uri"]) if code.get("uri") else None
            if media_type_uri:
                media_types[
                    remove_scheme_and_trailing_slash(media_type_uri)
                ] = MediaTypeOrExtent(
                    uri=code.get("uri"),
                    name=f"{code.get('name')}",
                    code=f"{code.get('type')}/{code.get('subType')}",
                    type=MediaTypeOrExtentType.MEDIA_TYPE,
                )

    file_types = get_reference_data("/eu/file-types").get("fileTypes")
    if file_types is not None:
        for file_type in file_types:
            file_type_uri = str(file_type["uri"]) if file_type.get("uri") else None
            if file_type_uri:
                media_types[
                    remove_scheme_and_trailing_slash(file_type_uri)
                ] = MediaTypeOrExtent(
                    uri=file_type.get("uri"),
                    name=f"{file_type.get('code')}",
                    code=f"{file_type.get('code')}",
                    type=MediaTypeOrExtentType.FILE_TYPE,
                )

    return media_types if len(media_types) > 0 else None



from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    RDF,
    SKOS,
    XSD,
)

from fdk_rdf_parser.classes.model_element import (
    ModelCodeElement,
    ModelElement,
)
from fdk_rdf_parser.rdf_utils import (
    model_dcat_ap_no_uri,
    object_value,
    resource_list,
    uri_or_identifier_list,
    value_set,
    value_translations,
    xkos_uri,
)

CODE_LIST = model_dcat_ap_no_uri("CodeList").toPython()
SIMPLE_TYPE = model_dcat_ap_no_uri("SimpleType").toPython()


def parse_model_element(graph: Graph, element_ref: URIRef) -> ModelElement:

    element_uri = None
    if isinstance(element_ref, URIRef):
        element_uri = element_ref.toPython()

    has_property_refs = resource_list(
        graph, element_ref, model_dcat_ap_no_uri("hasProperty")
    )

    element = ModelElement(
        uri=element_uri,
        identifier=object_value(graph, element_ref, DCTERMS.identifier),
        elementTypes=value_set(graph, element_ref, RDF.type),
        title=value_translations(graph, element_ref, DCTERMS.title),
        description=value_translations(graph, element_ref, DCTERMS.description),
        subject=object_value(graph, element_ref, DCTERMS.subject),
        belongsToModule=object_value(
            graph, element_ref, model_dcat_ap_no_uri("belongsToModule")
        ),
        hasProperty=uri_or_identifier_list(graph, has_property_refs),
    )

    if element.elementTypes is None:
        return element

    if CODE_LIST in element.elementTypes:
        return parse_model_code_list(graph, element_ref, element)
    elif SIMPLE_TYPE in element.elementTypes:
        return parse_model_simple_type(graph, element_ref, element)
    else:
        return element


def parse_model_code_list(
    graph: Graph, element_ref: URIRef, element: ModelElement
) -> ModelElement:
    codes = []
    for code_ref in graph.subjects(SKOS.inScheme, element_ref):
        codes.append(
            ModelCodeElement(
                uri=code_ref.toPython() if isinstance(code_ref, URIRef) else None,
                identifier=object_value(graph, code_ref, DCTERMS.identifier),
                prefLabel=value_translations(graph, code_ref, SKOS.prefLabel),
                inScheme=value_set(graph, code_ref, SKOS.inScheme),
                subject=object_value(graph, code_ref, DCTERMS.subject),
                notation=object_value(graph, code_ref, SKOS.notation),
                topConceptOf=value_set(graph, code_ref, SKOS.topConceptOf),
                definition=value_set(graph, code_ref, SKOS.definition),
                example=value_set(graph, code_ref, SKOS.example),
                exclusionNote=value_translations(
                    graph, code_ref, xkos_uri("exclusionNote")
                ),
                previousElement=value_set(graph, code_ref, xkos_uri("previous")),
                hiddenLabel=value_translations(graph, code_ref, SKOS.hiddenLabel),
                inclusionNote=value_translations(
                    graph, code_ref, xkos_uri("inclusionNote")
                ),
                note=value_translations(graph, code_ref, SKOS.note),
                nextElement=value_set(graph, code_ref, xkos_uri("next")),
                scopeNote=value_translations(graph, code_ref, SKOS.scopeNote),
                altLabel=value_translations(graph, code_ref, SKOS.altLabel),
            )
        )
    codes.sort()

    element.codeListReference = object_value(
        graph, element_ref, model_dcat_ap_no_uri("codeListReference")
    )
    element.codes = codes

    return element


def parse_model_simple_type(
    graph: Graph, element_ref: URIRef, element: ModelElement
) -> ModelElement:

    element.typeDefinitionReference = object_value(
        graph, element_ref, model_dcat_ap_no_uri("typeDefinitionReference")
    )
    element.fractionDigits = object_value(graph, element_ref, XSD.fractionDigits)
    element.length = object_value(graph, element_ref, XSD.length)
    element.maxInclusive = object_value(graph, element_ref, XSD.maxInclusive)
    element.maxLength = object_value(graph, element_ref, XSD.maxLength)
    element.minInclusive = object_value(graph, element_ref, XSD.minInclusive)
    element.minLength = object_value(graph, element_ref, XSD.minLength)
    element.pattern = object_value(graph, element_ref, XSD.pattern)
    element.totalDigits = object_value(graph, element_ref, XSD.totalDigits)

    return element



from dataclasses import dataclass
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Set,
)

from .catalog import Catalog
from .conforms_to import ConformsTo
from .dcat_resource import PartialDcatResource
from .distribution import Distribution
from .harvest_meta_data import HarvestMetaData
from .publisher import Publisher
from .qualified_attribution import QualifiedAttribution
from .quality_annotation import QualityAnnotation
from .references import Reference
from .skos_code import ReferenceDataCode
from .skos_concept import SkosConcept
from .subject import Subject
from .temporal import Temporal
from .theme import LosNode


@dataclass
class PartialDataset(PartialDcatResource):
    id: Optional[str] = None
    harvest: Optional[HarvestMetaData] = None
    accessRightsComment: Optional[List[str]] = None
    distribution: Optional[List[Distribution]] = None
    sample: Optional[List[Distribution]] = None
    source: Optional[str] = None
    objective: Optional[Dict[str, str]] = None
    page: Optional[Set[str]] = None
    admsIdentifier: Optional[Set[str]] = None
    temporal: Optional[List[Temporal]] = None
    subject: Optional[List[Subject]] = None
    spatial: Optional[List[ReferenceDataCode]] = None
    provenance: Optional[ReferenceDataCode] = None
    accrualPeriodicity: Optional[ReferenceDataCode] = None
    hasAccuracyAnnotation: Optional[QualityAnnotation] = None
    hasCompletenessAnnotation: Optional[QualityAnnotation] = None
    hasCurrentnessAnnotation: Optional[QualityAnnotation] = None
    hasAvailabilityAnnotation: Optional[QualityAnnotation] = None
    hasRelevanceAnnotation: Optional[QualityAnnotation] = None
    legalBasisForRestriction: Optional[List[SkosConcept]] = None
    legalBasisForProcessing: Optional[List[SkosConcept]] = None
    legalBasisForAccess: Optional[List[SkosConcept]] = None
    conformsTo: Optional[List[ConformsTo]] = None
    informationModel: Optional[List[SkosConcept]] = None
    references: Optional[List[Reference]] = None
    qualifiedAttributions: Optional[List[QualifiedAttribution]] = None
    catalog: Optional[Catalog] = None
    isOpenData: bool = False
    isAuthoritative: bool = False
    isRelatedToTransportportal: bool = False
    inSeries: Optional[str] = None
    prev: Optional[str] = None

    def add_values_from_dcat_resource(self: Any, values: PartialDcatResource) -> None:
        self.identifier = values.identifier
        self.publisher = values.publisher
        self.title = values.title
        self.description = values.description
        self.descriptionFormatted = values.descriptionFormatted
        self.uri = values.uri
        self.accessRights = values.accessRights
        self.theme = values.theme
        self.keyword = values.keyword
        self.contactPoint = values.contactPoint
        self.dctType = values.dctType
        self.issued = values.issued
        self.modified = values.modified
        self.landingPage = values.landingPage
        self.language = values.language


@dataclass
class Dataset(PartialDataset):
    publisher: Optional[Publisher] = None
    losTheme: Optional[List[LosNode]] = None
    type: str = "datasets"

    def add_values_from_partial(self: Any, values: PartialDataset) -> None:
        self.id = values.id
        self.identifier = values.identifier
        self.title = values.title
        self.publisher = values.publisher
        self.description = values.description
        self.descriptionFormatted = values.descriptionFormatted
        self.uri = values.uri
        self.accessRights = values.accessRights
        self.theme = values.theme
        self.keyword = values.keyword
        self.contactPoint = values.contactPoint
        self.dctType = values.dctType
        self.issued = values.issued
        self.modified = values.modified
        self.landingPage = values.landingPage
        self.language = values.language
        self.harvest = values.harvest
        self.accessRightsComment = values.accessRightsComment
        self.distribution = values.distribution
        self.sample = values.sample
        self.source = values.source
        self.objective = values.objective
        self.page = values.page
        self.admsIdentifier = values.admsIdentifier
        self.temporal = values.temporal
        self.subject = values.subject
        self.spatial = values.spatial
        self.provenance = values.provenance
        self.accrualPeriodicity = values.accrualPeriodicity
        self.hasAccuracyAnnotation = values.hasAccuracyAnnotation
        self.hasCompletenessAnnotation = values.hasCompletenessAnnotation
        self.hasCurrentnessAnnotation = values.hasCurrentnessAnnotation
        self.hasAvailabilityAnnotation = values.hasAvailabilityAnnotation
        self.hasRelevanceAnnotation = values.hasRelevanceAnnotation
        self.legalBasisForRestriction = values.legalBasisForRestriction
        self.legalBasisForProcessing = values.legalBasisForProcessing
        self.legalBasisForAccess = values.legalBasisForAccess
        self.conformsTo = values.conformsTo
        self.informationModel = values.informationModel
        self.references = values.references
        self.qualifiedAttributions = values.qualifiedAttributions
        self.catalog = values.catalog
        self.isOpenData = values.isOpenData
        self.isAuthoritative = values.isAuthoritative
        self.isRelatedToTransportportal = values.isRelatedToTransportportal
        self.inSeries = values.inSeries
        self.prev = values.prev


@dataclass
class DatasetSeries(Dataset):
    first: Optional[str] = None
    last: Optional[str] = None
    specialized_type: str = "dataset_series"

    def add_values_from_dataset(self: Any, values: Dataset) -> None:
        self.id = values.id
        self.identifier = values.identifier
        self.title = values.title
        self.publisher = values.publisher
        self.description = values.description
        self.descriptionFormatted = values.descriptionFormatted
        self.uri = values.uri
        self.accessRights = values.accessRights
        self.theme = values.theme
        self.keyword = values.keyword
        self.contactPoint = values.contactPoint
        self.dctType = values.dctType
        self.issued = values.issued
        self.modified = values.modified
        self.landingPage = values.landingPage
        self.language = values.language
        self.harvest = values.harvest
        self.accessRightsComment = values.accessRightsComment
        self.distribution = values.distribution
        self.sample = values.sample
        self.source = values.source
        self.objective = values.objective
        self.page = values.page
        self.admsIdentifier = values.admsIdentifier
        self.temporal = values.temporal
        self.subject = values.subject
        self.spatial = values.spatial
        self.provenance = values.provenance
        self.accrualPeriodicity = values.accrualPeriodicity
        self.hasAccuracyAnnotation = values.hasAccuracyAnnotation
        self.hasCompletenessAnnotation = values.hasCompletenessAnnotation
        self.hasCurrentnessAnnotation = values.hasCurrentnessAnnotation
        self.hasAvailabilityAnnotation = values.hasAvailabilityAnnotation
        self.hasRelevanceAnnotation = values.hasRelevanceAnnotation
        self.legalBasisForRestriction = values.legalBasisForRestriction
        self.legalBasisForProcessing = values.legalBasisForProcessing
        self.legalBasisForAccess = values.legalBasisForAccess
        self.conformsTo = values.conformsTo
        self.informationModel = values.informationModel
        self.references = values.references
        self.qualifiedAttributions = values.qualifiedAttributions
        self.catalog = values.catalog
        self.isOpenData = values.isOpenData
        self.isAuthoritative = values.isAuthoritative
        self.isRelatedToTransportportal = values.isRelatedToTransportportal
        self.inSeries = values.inSeries
        self.prev = values.prev
        self.publisher = values.publisher
        self.losTheme = values.losTheme
        self.type = values.type



from .ns import (
    adms_uri,
    br_uri,
    cpsv_uri,
    cpsvno_uri,
    cv_uri,
    dcat_ap_no_uri,
    dcat_api_uri,
    dcat_uri,
    dct_uri,
    dqv_iso_uri,
    dqv_uri,
    fdk_uri,
    model_dcat_ap_no_uri,
    oa_uri,
    owl_time_uri,
    prof_uri,
    prov_uri,
    rov_uri,
    schema_uri,
    skosno_old_uri,
    skosno_uri,
    skosxl_uri,
    vcard_uri,
    xkos_uri,
    xkos_uri_v_2,
)
from .utils import (
    date_list,
    date_value,
    duration_string_value,
    is_type,
    is_uri_in_list,
    linguistic_system_keywords,
    object_number_value,
    object_value,
    resource_list,
    string_to_float,
    uri_or_identifier,
    uri_or_identifier_list,
    value_list,
    value_set,
    value_translations,
)



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    FOAF,
    RDF,
)

from fdk_rdf_parser.classes import Agent
from fdk_rdf_parser.rdf_utils import (
    cv_uri,
    is_uri_in_list,
    object_value,
    value_list,
    value_translations,
)


def extract_agents_for_participation(
    graph: Graph, participation_ref: URIRef
) -> Optional[List[Agent]]:
    values = []
    agents = graph.subjects(predicate=RDF.type, object=DCTERMS.Agent)
    filtered_agents = filter(
        lambda agent: is_uri_in_list(
            participation_ref, graph, agent, cv_uri("playsRole")
        ),
        agents,
    )
    for resource in filtered_agents:
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None

        values.append(
            Agent(
                uri=resource_uri,
                identifier=object_value(graph, resource, DCTERMS.identifier),
                title=value_translations(graph, resource, DCTERMS.title),
                name=object_value(graph, resource, FOAF.name),
                playsRole=value_list(graph, resource, cv_uri("playsRole")),
            )
        )

    return values if len(values) > 0 else None



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    SKOS,
)

from fdk_rdf_parser.classes import Subject
from fdk_rdf_parser.rdf_utils import (
    object_value,
    resource_list,
    value_translations,
)


def extract_subjects(graph: Graph, dataset_ref: URIRef) -> Optional[List[Subject]]:
    dataset_subjects = []
    for subject_ref in resource_list(graph, dataset_ref, DCTERMS.subject):
        dataset_subjects.append(
            Subject(
                uri=subject_ref.toPython() if isinstance(subject_ref, URIRef) else None,
                identifier=object_value(graph, subject_ref, DCTERMS.identifier),
                prefLabel=value_translations(graph, subject_ref, SKOS.prefLabel),
                definition=value_translations(graph, subject_ref, SKOS.definition),
            )
        )

    return dataset_subjects if len(dataset_subjects) > 0 else None



from dataclasses import dataclass
from typing import (
    Dict,
    Optional,
)


@dataclass
class QualityAnnotation:
    inDimension: Optional[str] = None
    hasBody: Optional[Dict[str, str]] = None



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    FOAF,
)

from fdk_rdf_parser.classes import (
    PublicService,
    Service,
)
from fdk_rdf_parser.rdf_utils import (
    adms_uri,
    cpsv_uri,
    cv_uri,
    duration_string_value,
    is_type,
    object_value,
    resource_list,
    value_list,
    value_translations,
)
from .channel import extract_channels
from .contactpoint import extract_cv_has_contact_point
from .cost import extract_costs
from .dcat_resource import (
    extract_key_words,
    extract_skos_code,
    extract_skos_code_list,
)
from .evidence import extract_evidences
from .harvest_meta_data import extract_meta_data
from .legal_resource import extract_legal_resources
from .output import extract_outputs
from .participation import extract_participations
from .publisher import extract_list_of_publishers
from .requirement import extract_requirements
from .rule import extract_rules
from .skos_concept import extract_skos_concept


def extract_cpsvno_services(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[Service]]:
    values = []
    for resource in resource_list(graph, subject, predicate):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None
        values.append(
            Service(
                uri=resource_uri,
                identifier=object_value(graph, resource, DCTERMS.identifier),
                id=object_value(graph, resource, DCTERMS.identifier),
                title=value_translations(graph, resource, DCTERMS.title),
            )
        )

    return values if len(values) > 0 else None


def parse_cpsvno_service(
    services_graph: Graph, catalog_record_uri: URIRef, cpsvno_service_uri: URIRef
) -> Service:

    service = Service(
        id=object_value(services_graph, catalog_record_uri, DCTERMS.identifier),
        uri=cpsvno_service_uri.toPython(),
        identifier=object_value(services_graph, cpsvno_service_uri, DCTERMS.identifier),
        title=value_translations(services_graph, cpsvno_service_uri, DCTERMS.title),
        description=value_translations(
            services_graph, cpsvno_service_uri, DCTERMS.description
        ),
        ownedBy=extract_list_of_publishers(
            services_graph, cpsvno_service_uri, cv_uri("ownedBy")
        ),
        contactPoint=extract_cv_has_contact_point(services_graph, cpsvno_service_uri),
        isGroupedBy=value_list(
            services_graph, cpsvno_service_uri, cv_uri("isGroupedBy")
        ),
        harvest=extract_meta_data(services_graph, catalog_record_uri),
        keyword=extract_key_words(services_graph, cpsvno_service_uri),
        sector=extract_skos_concept(
            services_graph, cpsvno_service_uri, cv_uri("sector")
        ),
        isClassifiedBy=extract_skos_concept(
            services_graph, cpsvno_service_uri, cv_uri("isClassifiedBy")
        ),
        isDescribedAt=extract_skos_concept(
            services_graph, cpsvno_service_uri, cv_uri("isDescribedAt")
        ),
        language=extract_skos_code_list(
            services_graph, cpsvno_service_uri, DCTERMS.language
        ),
        holdsRequirement=extract_requirements(services_graph, cpsvno_service_uri),
        hasParticipation=extract_participations(services_graph, cpsvno_service_uri),
        hasInput=extract_evidences(services_graph, cpsvno_service_uri),
        produces=extract_outputs(services_graph, cpsvno_service_uri),
        requires=extract_cpsvno_services(
            services_graph, cpsvno_service_uri, DCTERMS.requires
        ),
        follows=extract_rules(services_graph, cpsvno_service_uri),
        hasLegalResource=extract_legal_resources(
            services_graph, cpsvno_service_uri, cv_uri("hasLegalResource")
        ),
        hasChannel=extract_channels(
            services_graph, cpsvno_service_uri, cv_uri("hasChannel")
        ),
        processingTime=duration_string_value(
            services_graph, cpsvno_service_uri, cv_uri("processingTime")
        ),
        hasCost=extract_costs(services_graph, cpsvno_service_uri),
        relation=extract_cpsvno_services(
            services_graph, cpsvno_service_uri, DCTERMS.relation
        ),
        spatial=value_list(services_graph, cpsvno_service_uri, DCTERMS.spatial),
        admsStatus=extract_skos_code(
            services_graph, cpsvno_service_uri, adms_uri("status")
        ),
        subject=value_list(services_graph, cpsvno_service_uri, DCTERMS.subject),
        homepage=value_list(services_graph, cpsvno_service_uri, FOAF.homepage),
        dctType=extract_skos_code_list(
            services_graph, cpsvno_service_uri, DCTERMS.type
        ),
        thematicAreaUris=value_list(
            services_graph, cpsvno_service_uri, cv_uri("thematicArea")
        ),
    )

    if is_type(cpsv_uri("PublicService"), services_graph, cpsvno_service_uri):
        public_service = PublicService(
            hasCompetentAuthority=extract_list_of_publishers(
                services_graph, cpsvno_service_uri, cv_uri("hasCompetentAuthority")
            ),
        )
        public_service.add_cpsvno_service_values(service)
        return public_service
    else:
        return service



from dataclasses import dataclass
from typing import (
    List,
    Optional,
)

from .skos_code import ReferenceDataCode


@dataclass
class OpeningHoursSpecification:
    uri: Optional[str] = None
    dayOfWeek: Optional[List[ReferenceDataCode]] = None
    opens: Optional[str] = None
    closes: Optional[str] = None
    validFrom: Optional[str] = None
    validThrough: Optional[str] = None



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)

from .skos_code import ReferenceDataCode
from .skos_concept import SkosConcept


@dataclass
class Output:
    uri: Optional[str] = None
    identifier: Optional[str] = None
    name: Optional[Dict[str, str]] = None
    description: Optional[Dict[str, str]] = None
    language: Optional[List[ReferenceDataCode]] = None
    type: Optional[List[SkosConcept]] = None



from typing import (
    Dict,
    List,
    Optional,
)

from fdk_rdf_parser.classes import (
    DataService,
    MediaTypeOrExtent,
    ReferenceDataCode,
)
from .reference_data import DataServiceReferenceData
from .utils import remove_scheme_and_trailing_slash


def extend_data_service_with_reference_data(
    data_service: DataService, ref_data: DataServiceReferenceData
) -> DataService:
    data_service.mediaType = extend_media_type_skos_codes(
        data_service.mediaType, ref_data.media_types
    )
    data_service.fdkFormat = extend_fdk_format(
        data_service.fdkFormat, ref_data.media_types
    )

    return data_service


def extend_media_type_skos_codes(
    media_types: Optional[List[ReferenceDataCode]],
    references: Optional[Dict[str, MediaTypeOrExtent]],
) -> Optional[List[ReferenceDataCode]]:
    extended = []
    if media_types and references:
        for media_type in media_types:
            reference = find_corresponding_media_type_reference(
                media_type.uri, references
            )
            if reference:
                extended.append(
                    ReferenceDataCode(
                        uri=media_type.uri,
                        code=reference.code,
                        prefLabel={"nb": reference.code} if reference.code else None,
                    )
                )
    return extended if len(extended) > 0 else None


def extend_fdk_format(
    formats: Optional[List[MediaTypeOrExtent]],
    references: Optional[Dict[str, MediaTypeOrExtent]],
) -> Optional[List[MediaTypeOrExtent]]:
    extended = []
    if formats and references:
        for fmt in formats:
            reference = find_corresponding_media_type_reference(fmt.uri, references)
            if reference:
                extended.append(reference)
            elif fmt.code:
                extended.append(fmt)
    return extended if len(extended) > 0 else None


def find_corresponding_media_type_reference(
    media_type: Optional[str], references: Optional[Dict[str, MediaTypeOrExtent]]
) -> Optional[MediaTypeOrExtent]:
    if media_type and references:
        return references.get(remove_scheme_and_trailing_slash(media_type))
    return None



from .agent import Agent
from .business_event import BusinessEvent
from .catalog import Catalog
from .channel import Channel
from .concept import Concept
from .conforms_to import ConformsTo
from .contactpoint import (
    CVContactPoint,
    DCATContactPoint,
)
from .cost import Cost
from .cpsvno_service import Service
from .data_distribution_service import DataDistributionService
from .dataservice import DataService
from .dataset import (
    Dataset,
    DatasetSeries,
    PartialDataset,
)
from .dcat_resource import PartialDcatResource
from .dct_standard import DctStandard
from .distribution import Distribution
from .event import Event
from .evidence import Evidence
from .format import Format
from .harvest_meta_data import HarvestMetaData
from .info_model import InformationModel
from .legal_resource import LegalResource
from .life_event import LifeEvent
from .media_type import (
    MediaTypeOrExtent,
    MediaTypeOrExtentType,
)
from .opening_hours_specification import OpeningHoursSpecification
from .output import Output
from .participation import Participation
from .public_service import PublicService
from .publisher import Publisher
from .qualified_attribution import QualifiedAttribution
from .quality_annotation import QualityAnnotation
from .references import Reference
from .requirement import Requirement
from .rule import Rule
from .skos_code import ReferenceDataCode
from .skos_concept import SkosConcept
from .subject import Subject
from .temporal import Temporal
from .theme import (
    ConceptSchema,
    EuDataTheme,
    LosNode,
)



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
    Set,
)

from .conforms_to import ConformsTo
from .data_distribution_service import DataDistributionService
from .media_type import MediaTypeOrExtent
from .skos_code import ReferenceDataCode
from .skos_concept import SkosConcept


@dataclass
class Distribution:
    uri: Optional[str] = None
    title: Optional[Dict[str, str]] = None
    description: Optional[Dict[str, str]] = None
    downloadURL: Optional[Set[str]] = None
    accessURL: Optional[Set[str]] = None
    license: Optional[List[SkosConcept]] = None
    openLicense: bool = False
    conformsTo: Optional[List[ConformsTo]] = None
    page: Optional[List[SkosConcept]] = None
    mediaType: Optional[List[ReferenceDataCode]] = None
    format: Optional[Set[str]] = None
    fdkFormat: Optional[List[MediaTypeOrExtent]] = None
    compressFormat: Optional[MediaTypeOrExtent] = None
    packageFormat: Optional[MediaTypeOrExtent] = None
    type: Optional[str] = None
    accessService: Optional[List[DataDistributionService]] = None



from dataclasses import dataclass
from typing import Optional

from .publisher import Publisher


@dataclass
class QualifiedAttribution:
    agent: Optional[Publisher] = None
    role: Optional[str] = None



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)

from .channel import Channel
from .publisher import Publisher


@dataclass
class Cost:
    uri: Optional[str] = None
    identifier: Optional[str] = None
    description: Optional[Dict[str, str]] = None
    currency: Optional[str] = None
    ifAccessedThrough: Optional[Channel] = None
    isDefinedBy: Optional[List[Publisher]] = None
    value: Optional[str] = None



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)

from fdk_rdf_parser.classes import OpeningHoursSpecification
from fdk_rdf_parser.parse_functions.skos_code import extract_skos_code_list
from fdk_rdf_parser.rdf_utils import (
    date_value,
    resource_list,
    schema_uri,
)


def extract_opening_hours_specification(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[OpeningHoursSpecification]]:
    values = []
    for resource in resource_list(graph, subject, predicate):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None
        values.append(
            OpeningHoursSpecification(
                uri=resource_uri,
                dayOfWeek=extract_skos_code_list(
                    graph, resource, schema_uri("dayOfWeek")
                ),
                opens=date_value(graph, resource, schema_uri("opens")),
                closes=date_value(graph, resource, schema_uri("closes")),
                validFrom=date_value(graph, resource, schema_uri("validFrom")),
                validThrough=date_value(graph, resource, schema_uri("validThrough")),
            )
        )

    return values if len(values) > 0 else None



from dataclasses import dataclass
from typing import (
    Dict,
    Optional,
)


@dataclass
class Subject:
    identifier: Optional[str] = None
    uri: Optional[str] = None
    prefLabel: Optional[Dict[str, str]] = None
    definition: Optional[Dict[str, str]] = None



from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import HarvestMetaData
from fdk_rdf_parser.rdf_utils import (
    date_list,
    date_value,
)


def extract_meta_data(graph: Graph, subject: URIRef) -> HarvestMetaData:
    return HarvestMetaData(
        firstHarvested=date_value(graph, subject, DCTERMS.issued),
        changed=date_list(graph, subject, DCTERMS.modified),
    )



from typing import (
    Dict,
    List,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    FOAF,
    RDFS,
    SKOS,
)

from fdk_rdf_parser.classes import (
    DatasetSeries,
    PartialDataset,
    SkosConcept,
)
from fdk_rdf_parser.rdf_utils import (
    adms_uri,
    cpsv_uri,
    dcat_ap_no_uri,
    dcat_uri,
    dqv_iso_uri,
    fdk_uri,
    object_value,
    prov_uri,
    resource_list,
    value_list,
    value_set,
    value_translations,
)
from .catalog import parse_catalog
from .conforms_to import extract_conforms_to
from .dcat_resource import parse_dcat_resource
from .distribution import extract_distributions
from .harvest_meta_data import extract_meta_data
from .qualified_attribution import extract_qualified_attributions
from .quality_annotation import extract_quality_annotation
from .references import extract_references
from .skos_code import (
    extract_skos_code,
    extract_skos_code_list,
)
from .skos_concept import extract_skos_concept
from .subject import extract_subjects
from .temporal import extract_temporal


def parse_dataset(
    datasets_graph: Graph, record_uri: URIRef, dataset_uri: URIRef
) -> PartialDataset:
    quality_annotations = extract_quality_annotation(datasets_graph, dataset_uri)
    cpsv_follows = extract_legal_basis_from_cpsv_follows(datasets_graph, dataset_uri)

    dataset = PartialDataset(
        id=object_value(datasets_graph, record_uri, DCTERMS.identifier),
        admsIdentifier=value_set(datasets_graph, dataset_uri, adms_uri("identifier")),
        harvest=extract_meta_data(datasets_graph, record_uri),
        accessRightsComment=value_list(
            datasets_graph, dataset_uri, dcat_ap_no_uri("accessRightsComment")
        ),
        distribution=extract_distributions(
            datasets_graph, dataset_uri, dcat_uri("distribution")
        ),
        sample=extract_distributions(datasets_graph, dataset_uri, adms_uri("sample")),
        spatial=extract_skos_code_list(datasets_graph, dataset_uri, DCTERMS.spatial),
        source=object_value(datasets_graph, dataset_uri, dcat_ap_no_uri("source")),
        objective=value_translations(
            datasets_graph, dataset_uri, dcat_ap_no_uri("objective")
        ),
        page=value_set(datasets_graph, dataset_uri, FOAF.page),
        temporal=extract_temporal(datasets_graph, dataset_uri),
        subject=extract_subjects(datasets_graph, dataset_uri),
        provenance=extract_skos_code(datasets_graph, dataset_uri, DCTERMS.provenance),
        accrualPeriodicity=extract_skos_code(
            datasets_graph, dataset_uri, DCTERMS.accrualPeriodicity
        ),
        hasAccuracyAnnotation=quality_annotations.get(
            dqv_iso_uri("Accuracy").toPython()
        ),
        hasCompletenessAnnotation=quality_annotations.get(
            dqv_iso_uri("Completeness").toPython()
        ),
        hasCurrentnessAnnotation=quality_annotations.get(
            dqv_iso_uri("Currentness").toPython()
        ),
        hasAvailabilityAnnotation=quality_annotations.get(
            dqv_iso_uri("Availability").toPython()
        ),
        hasRelevanceAnnotation=quality_annotations.get(
            dqv_iso_uri("Relevance").toPython()
        ),
        legalBasisForRestriction=cpsv_follows["restriction"]
        if len(cpsv_follows["restriction"]) > 0
        else extract_skos_concept(
            datasets_graph, dataset_uri, dcat_ap_no_uri("legalBasisForRestriction")
        ),
        legalBasisForProcessing=cpsv_follows["processing"]
        if len(cpsv_follows["processing"]) > 0
        else extract_skos_concept(
            datasets_graph, dataset_uri, dcat_ap_no_uri("legalBasisForProcessing")
        ),
        legalBasisForAccess=cpsv_follows["access"]
        if len(cpsv_follows["access"]) > 0
        else extract_skos_concept(
            datasets_graph, dataset_uri, dcat_ap_no_uri("legalBasisForAccess")
        ),
        conformsTo=extract_conforms_to(datasets_graph, dataset_uri),
        informationModel=extract_skos_concept(
            datasets_graph, dataset_uri, dcat_ap_no_uri("informationModel")
        ),
        references=extract_references(datasets_graph, dataset_uri),
        qualifiedAttributions=extract_qualified_attributions(
            datasets_graph, dataset_uri, prov_uri("qualifiedAttribution")
        ),
        catalog=parse_catalog(datasets_graph, record_uri),
        isOpenData=extract_boolean(datasets_graph, dataset_uri, fdk_uri("isOpenData")),
        isAuthoritative=extract_boolean(
            datasets_graph, dataset_uri, fdk_uri("isAuthoritative")
        ),
        isRelatedToTransportportal=extract_boolean(
            datasets_graph, dataset_uri, fdk_uri("isRelatedToTransportportal")
        ),
        inSeries=object_value(datasets_graph, dataset_uri, dcat_uri("inSeries")),
        prev=object_value(datasets_graph, dataset_uri, dcat_uri("prev")),
    )

    dataset.add_values_from_dcat_resource(
        parse_dcat_resource(
            graph=datasets_graph,
            subject=dataset_uri,
        )
    )

    return dataset


def parse_dataset_series_values(
    datasets_graph: Graph, dataset_uri: URIRef
) -> DatasetSeries:
    return DatasetSeries(
        last=object_value(datasets_graph, dataset_uri, dcat_uri("last")),
        first=object_value(datasets_graph, dataset_uri, dcat_uri("first")),
    )


def extract_boolean(
    datasets_graph: Graph, dataset_uri: URIRef, predicate: URIRef
) -> bool:
    value = object_value(datasets_graph, dataset_uri, predicate)
    return value if value and isinstance(value, bool) else False


def extract_legal_basis_from_cpsv_follows(
    datasets_graph: Graph, dataset_uri: URIRef
) -> Dict[str, List[SkosConcept]]:
    legal_basis_for: Dict[str, List[SkosConcept]] = {
        "restriction": [],
        "processing": [],
        "access": [],
    }

    for legal_resource in resource_list(
        datasets_graph, dataset_uri, cpsv_uri("follows")
    ):
        rule_type = object_value(datasets_graph, legal_resource, DCTERMS.type)
        implements_ref = datasets_graph.value(legal_resource, cpsv_uri("implements"))
        if rule_type and implements_ref:
            legal_type_ref = datasets_graph.value(implements_ref, DCTERMS.type)
            skos_concept = SkosConcept(
                uri=object_value(datasets_graph, implements_ref, RDFS.seeAlso),
                extraType=rule_type,
                prefLabel=value_translations(
                    datasets_graph, legal_type_ref, SKOS.prefLabel
                )
                if legal_type_ref
                else None,
            )

            if "ruleForNonDisclosure" in rule_type:
                legal_basis_for["restriction"].append(skos_concept)
            elif "ruleForDataProcessing" in rule_type:
                legal_basis_for["processing"].append(skos_concept)
            elif "ruleForDisclosure" in rule_type:
                legal_basis_for["access"].append(skos_concept)

    return legal_basis_for



import logging
from typing import (
    Dict,
    List,
)

import requests

from .utils import reference_data_url


def get_reference_data(endpoint: str) -> Dict[str, List[Dict]]:

    try:
        response = requests.get(
            reference_data_url(endpoint), headers={"Accept": "application/json"}
        )

        response.raise_for_status()

        return response.json()

    except requests.HTTPError as err:
        logging.error(f"Http error response from reference-data ({err})")
    except Exception as err:
        logging.error(f"Error occurred when getting data from reference-data: {err}")

    return {}



from fdk_rdf_parser.classes import InformationModel
from .reference_data import InformationModelReferenceData
from .utils import (
    extend_eu_data_themes,
    extend_los_themes,
    extend_skos_code_list,
    split_themes,
)


def extend_info_model_with_reference_data(
    info_model: InformationModel, ref_data: InformationModelReferenceData
) -> InformationModel:
    info_model.language = extend_skos_code_list(
        info_model.language, ref_data.linguisticsystem
    )
    info_model.spatial = extend_skos_code_list(info_model.spatial, ref_data.location)
    info_model.license = extend_skos_code_list(
        info_model.license, ref_data.openlicenses
    )

    splitted_themes = split_themes(info_model.theme, ref_data.los_themes)
    info_model.losTheme = extend_los_themes(splitted_themes["los"], ref_data.los_themes)
    info_model.theme = extend_eu_data_themes(
        splitted_themes["eu_data_themes"], ref_data.eu_data_themes
    )

    return info_model



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import Channel
from fdk_rdf_parser.rdf_utils import (
    object_value,
    resource_list,
)
from .skos_concept import extract_skos_concept


def extract_channels(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[Channel]]:
    values = []
    for resource in resource_list(graph, subject, predicate):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None
        skos_concepts = extract_skos_concept(graph, resource, DCTERMS.type)
        values.append(
            Channel(
                uri=resource_uri,
                identifier=object_value(graph, resource, DCTERMS.identifier),
                type=skos_concepts[0]
                if skos_concepts is not None and len(skos_concepts) == 1
                else None,
            )
        )

    return values if len(values) > 0 else None



from typing import Optional

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    FOAF,
)

from fdk_rdf_parser.classes import Catalog
from fdk_rdf_parser.rdf_utils import (
    dcat_uri,
    is_type,
    object_value,
    value_translations,
)
from .publisher import extract_publisher


def parse_catalog(graph: Graph, child_record_uri: URIRef) -> Optional[Catalog]:
    catalog_record_uri = graph.value(child_record_uri, DCTERMS.isPartOf)

    if catalog_record_uri and is_type(
        dcat_uri("CatalogRecord"), graph, catalog_record_uri
    ):
        catalog_uri = graph.value(catalog_record_uri, FOAF.primaryTopic)

        if catalog_uri and is_type(dcat_uri("Catalog"), graph, catalog_uri):

            return Catalog(
                id=object_value(graph, catalog_record_uri, DCTERMS.identifier),
                publisher=extract_publisher(graph, catalog_uri),
                title=value_translations(graph, catalog_uri, DCTERMS.title),
                uri=catalog_uri.toPython(),
                description=value_translations(graph, catalog_uri, DCTERMS.description),
            )
    return None



from typing import (
    Any,
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    RDFS,
)

from fdk_rdf_parser.classes import LegalResource
from fdk_rdf_parser.rdf_utils import (
    resource_list,
    value_list,
    value_translations,
)


def extract_legal_resources(
    graph: Graph, subject: URIRef, predicate: Any
) -> Optional[List[LegalResource]]:
    values = []
    for resource in resource_list(graph, subject, predicate):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None
        values.append(
            LegalResource(
                uri=resource_uri,
                dctTitle=value_translations(graph, resource, DCTERMS.title),
                description=value_translations(graph, resource, DCTERMS.description),
                seeAlso=value_list(graph, resource, RDFS.seeAlso),
                relation=value_list(graph, resource, DCTERMS.relation),
            )
        )

    return values if len(values) > 0 else None



from dataclasses import dataclass
from typing import Any

from .event import Event


@dataclass
class LifeEvent(Event):
    specialized_type: str = "life_event"

    def add_event_values(self: Any, values: Event) -> None:
        self.id = values.id
        self.uri = values.uri
        self.identifier = values.identifier
        self.harvest = values.harvest
        self.title = values.title
        self.description = values.description
        self.dctType = values.dctType
        self.hasCompetentAuthority = values.hasCompetentAuthority
        self.relation = values.relation
        self.associatedBroaderTypes = values.associatedBroaderTypes
        self.mayInitiate = values.mayInitiate
        self.subject = values.subject
        self.distribution = values.distribution



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    FOAF,
)

from fdk_rdf_parser.classes.evidence import (
    Evidence,
    EvidenceRdfType,
)
from fdk_rdf_parser.rdf_utils import (
    cpsv_uri,
    cv_uri,
    dcat_uri,
    is_type,
    object_value,
    resource_list,
    value_translations,
)
from fdk_rdf_parser.rdf_utils.utils import value_list
from .dcat_resource import extract_skos_code_list


def extract_evidences(graph: Graph, subject: URIRef) -> Optional[List[Evidence]]:
    values = []
    for resource in resource_list(graph, subject, cpsv_uri("hasInput")):
        resource_uri = resource.toPython() if isinstance(resource, URIRef) else None

        if is_type(cv_uri("Evidence"), graph, resource):
            rdf_type = EvidenceRdfType.EVIDENCE_TYPE.value
        elif is_type(dcat_uri("Dataset"), graph, resource):
            rdf_type = EvidenceRdfType.DATASET_TYPE.value
        else:
            rdf_type = EvidenceRdfType.UNKNOWN.value

        values.append(
            Evidence(
                rdfType=rdf_type,
                uri=resource_uri,
                identifier=object_value(graph, resource, DCTERMS.identifier),
                name=value_translations(graph, resource, DCTERMS.title),
                description=value_translations(graph, resource, DCTERMS.description),
                dctType=extract_skos_code_list(graph, resource, DCTERMS.type),
                language=extract_skos_code_list(graph, resource, DCTERMS.language),
                page=value_list(graph, resource, FOAF.page),
            )
        )

    return values if len(values) > 0 else None



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)

from fdk_rdf_parser.classes import QualifiedAttribution
from fdk_rdf_parser.rdf_utils import (
    dcat_uri,
    object_value,
    prov_uri,
    resource_list,
)
from .publisher import parse_publisher


def extract_qualified_attributions(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[QualifiedAttribution]]:
    values = []
    for resource in resource_list(graph, subject, predicate):
        publisher_uri = graph.value(resource, prov_uri("agent"))

        if isinstance(publisher_uri, URIRef):
            role = object_value(graph, resource, dcat_uri("hadRole"))

            values.append(
                QualifiedAttribution(
                    agent=parse_publisher(graph, publisher_uri), role=role
                )
            )

    return values if len(values) > 0 else None



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import DCTERMS

from fdk_rdf_parser.classes import DataDistributionService
from fdk_rdf_parser.rdf_utils import (
    dcat_api_uri,
    dcat_uri,
    resource_list,
    value_translations,
)
from .skos_concept import extract_skos_concept


def extract_data_distribution_services(
    graph: Graph, subject: URIRef
) -> Optional[List[DataDistributionService]]:
    values = []

    access_services = resource_list(graph, subject, dcat_api_uri("accessService"))
    access_services.extend(resource_list(graph, subject, dcat_uri("accessService")))

    for resource in access_services:
        resource_uri = None
        if isinstance(resource, URIRef):
            resource_uri = resource.toPython()

        endpoint_description = extract_skos_concept(
            graph, resource, dcat_uri("endpointDescription")
        )

        values.append(
            DataDistributionService(
                uri=resource_uri,
                title=value_translations(graph, resource, DCTERMS.title),
                description=value_translations(graph, resource, DCTERMS.description),
                endpointDescription=(
                    endpoint_description
                    if endpoint_description
                    else extract_skos_concept(
                        graph, resource, dcat_api_uri("endpointDescription")
                    )
                ),
            )
        )

    return values if len(values) > 0 else None



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)


@dataclass
class SkosConcept:
    uri: Optional[str] = None
    prefLabel: Optional[Dict[str, str]] = None
    extraType: Optional[str] = None
    broader: Optional[List[str]] = None
    narrower: Optional[List[str]] = None



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
)


@dataclass
class DctStandard:
    uri: Optional[str] = None
    title: Optional[Dict[str, str]] = None
    seeAlso: Optional[List[str]] = None
    versionInfo: Optional[str] = None



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)
from rdflib.namespace import (
    DCTERMS,
    RDFS,
)

from fdk_rdf_parser.classes import Format
from fdk_rdf_parser.rdf_utils import (
    dct_uri,
    object_value,
    resource_list,
    value_translations,
)
from .dcat_resource import extract_skos_code


def extract_formats(graph: Graph, subject: URIRef) -> Optional[List[Format]]:
    values = []
    for resource in resource_list(graph, subject, DCTERMS.hasFormat):
        format_uri = resource.toPython() if isinstance(resource, URIRef) else None
        values.append(
            Format(
                uri=format_uri,
                title=value_translations(graph, resource, DCTERMS.title),
                format=object_value(graph, resource, dct_uri("format")),
                language=extract_skos_code(graph, resource, DCTERMS.language),
                seeAlso=object_value(graph, resource, RDFS.seeAlso),
            )
        )

    return values if len(values) > 0 else None



from typing import (
    Dict,
    List,
    Optional,
    Set,
)

from fdk_rdf_parser.classes import (
    Dataset,
    Distribution,
    MediaTypeOrExtent,
    ReferenceDataCode,
)
from .reference_data import DatasetReferenceData
from .utils import (
    extend_eu_data_themes,
    extend_los_themes,
    extend_reference_types,
    extend_skos_code,
    extend_skos_code_list,
    remove_scheme_and_trailing_slash,
    split_themes,
)


def extend_dataset_with_reference_data(
    dataset: Dataset, ref_data: DatasetReferenceData
) -> Dataset:
    dataset.accessRights = extend_skos_code(
        dataset.accessRights, ref_data.rightsstatement
    )
    dataset.provenance = extend_skos_code(
        dataset.provenance, ref_data.provenancestatement
    )
    dataset.accrualPeriodicity = extend_skos_code(
        dataset.accrualPeriodicity, ref_data.frequency
    )
    dataset.language = extend_skos_code_list(
        dataset.language, ref_data.linguisticsystem
    )
    dataset.spatial = extend_skos_code_list(dataset.spatial, ref_data.location)
    dataset.distribution = extend_distributions(
        dataset.distribution, ref_data.openlicenses, ref_data.media_types
    )
    dataset.references = extend_reference_types(
        dataset.references, ref_data.referencetypes
    )

    splitted_themes = split_themes(dataset.theme, ref_data.los_themes)
    dataset.losTheme = extend_los_themes(splitted_themes["los"], ref_data.los_themes)
    dataset.theme = extend_eu_data_themes(
        splitted_themes["eu_data_themes"], ref_data.eu_data_themes
    )

    return dataset


def extend_distributions(
    distributions: Optional[List[Distribution]],
    open_licenses: Optional[Dict[str, ReferenceDataCode]],
    ref_media_types: Optional[Dict[str, MediaTypeOrExtent]],
) -> Optional[List[Distribution]]:
    if distributions is None:
        return distributions
    else:
        extended_distributions = []
        for dist in distributions:
            if open_licenses is not None:
                if dist.license is not None:
                    extended_licenses = []
                    for lic in dist.license:
                        ref_code = (
                            open_licenses.get(remove_scheme_and_trailing_slash(lic.uri))
                            if lic.uri is not None
                            else None
                        )
                        if ref_code is not None:
                            dist.openLicense = True
                            if lic.prefLabel is None:
                                lic.prefLabel = ref_code.prefLabel
                        extended_licenses.append(lic)
                    dist.license = extended_licenses

            if ref_media_types:
                fdk_formats: Set[MediaTypeOrExtent] = set()

                if dist.fdkFormat:
                    skos_code_formats = []
                    for fmt in dist.fdkFormat:
                        ref_media_type = (
                            ref_media_types.get(
                                remove_scheme_and_trailing_slash(fmt.uri)
                            )
                            if fmt.uri
                            else None
                        )
                        if ref_media_type:
                            skos_code_formats.append(
                                ReferenceDataCode(
                                    uri=ref_media_type.uri,
                                    code=ref_media_type.code,
                                    prefLabel={"nb": ref_media_type.code}
                                    if ref_media_type.code
                                    else None,
                                )
                            )
                            fdk_formats.add(ref_media_type)
                        elif fmt.code:
                            fdk_formats.add(fmt)
                    dist.fdkFormat = list(fdk_formats) if len(fdk_formats) > 0 else None
                    dist.mediaType = (
                        skos_code_formats if len(skos_code_formats) > 0 else None
                    )

                if (
                    dist.compressFormat
                    and dist.compressFormat.uri
                    and ref_media_types.get(
                        remove_scheme_and_trailing_slash(dist.compressFormat.uri)
                    )
                ):
                    dist.compressFormat = ref_media_types[
                        remove_scheme_and_trailing_slash(dist.compressFormat.uri)
                    ]

                if (
                    dist.packageFormat
                    and dist.packageFormat.uri
                    and ref_media_types.get(
                        remove_scheme_and_trailing_slash(dist.packageFormat.uri)
                    )
                ):
                    dist.packageFormat = ref_media_types[
                        remove_scheme_and_trailing_slash(dist.packageFormat.uri)
                    ]

            extended_distributions.append(dist)
        return extended_distributions



from dataclasses import dataclass
from typing import (
    Dict,
    List,
    Optional,
    Set,
)

from .contactpoint import DCATContactPoint
from .harvest_meta_data import HarvestMetaData
from .publisher import Publisher


@dataclass
class Collection:
    id: Optional[str] = None
    uri: Optional[str] = None
    label: Optional[Dict[str, str]] = None
    description: Optional[Dict[str, str]] = None
    publisher: Optional[Publisher] = None


@dataclass
class TextAndURI:
    uri: Optional[str] = None
    text: Optional[Dict[str, str]] = None


@dataclass
class Definition:
    text: Optional[Dict[str, str]] = None
    remark: Optional[Dict[str, str]] = None
    targetGroup: Optional[str] = None
    sourceRelationship: Optional[str] = None
    range: Optional[TextAndURI] = None
    sources: Optional[List[TextAndURI]] = None
    lastUpdated: Optional[str] = None


@dataclass
class AssociativeRelation:
    description: Optional[Dict[str, str]] = None
    related: Optional[str] = None


@dataclass
class PartitiveRelation:
    description: Optional[Dict[str, str]] = None
    hasPart: Optional[str] = None
    isPartOf: Optional[str] = None


@dataclass
class GenericRelation:
    divisioncriterion: Optional[Dict[str, str]] = None
    generalizes: Optional[str] = None
    specializes: Optional[str] = None


@dataclass
class Concept:
    id: Optional[str] = None
    uri: Optional[str] = None
    identifier: Optional[str] = None
    harvest: Optional[HarvestMetaData] = None
    collection: Optional[Collection] = None
    publisher: Optional[Publisher] = None
    subject: Optional[Dict[str, str]] = None
    application: Optional[List[Dict[str, str]]] = None
    example: Optional[Dict[str, str]] = None
    prefLabel: Optional[Dict[str, str]] = None
    hiddenLabel: Optional[List[Dict[str, str]]] = None
    altLabel: Optional[List[Dict[str, str]]] = None
    contactPoint: Optional[DCATContactPoint] = None
    definition: Optional[Definition] = None
    seeAlso: Optional[Set[str]] = None
    isReplacedBy: Optional[Set[str]] = None
    replaces: Optional[Set[str]] = None
    validFromIncluding: Optional[str] = None
    validToIncluding: Optional[str] = None
    associativeRelation: Optional[List[AssociativeRelation]] = None
    partitiveRelation: Optional[List[PartitiveRelation]] = None
    genericRelation: Optional[List[GenericRelation]] = None
    type: str = "concept"



from typing import (
    List,
    Optional,
)

from rdflib import (
    Graph,
    URIRef,
)

from fdk_rdf_parser.classes import ReferenceDataCode
from fdk_rdf_parser.rdf_utils import (
    object_value,
    value_list,
)


def extract_skos_code(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[ReferenceDataCode]:
    uri = object_value(graph, subject, predicate)
    return ReferenceDataCode(uri=uri) if uri is not None else None


def extract_skos_code_list(
    graph: Graph, subject: URIRef, predicate: URIRef
) -> Optional[List[ReferenceDataCode]]:
    skos_codes = []
    uri_list = value_list(graph, subject, predicate)
    if uri_list is None:
        return None
    else:
        for uri in uri_list:
            skos_codes.append(ReferenceDataCode(uri=uri))
        return skos_codes if len(skos_codes) > 0 else None
